﻿using Pk3DSRNGTool.RNG;

namespace Pk3DSRNGTool
{
    internal class ModelStatus
    {
        private SFMT sfmt;
        private int cnt;
        private ulong getrand { get { cnt++; return sfmt.Nextulong(); } }

        public byte Modelnumber;
        public int[] remain_frame;
        public bool phase;

        public bool IsBoy;
        public int fidget_cd = -1; // fidget cooldown. -1 to ignore fidget

        public bool raining;

        public ModelStatus(byte n, SFMT st)
        {
            sfmt = new SFMT(st);
            Modelnumber = n;
            remain_frame = new int[n];
        }

        public int NextState()
        {
            cnt = 0;
            if (fidget_cd > 0 && --fidget_cd == 0)
                fidget_cd = fidget();
            for (int i = 0; i < Modelnumber; i++)
            {
                if (remain_frame[i] > 1)                       //Cooldown 2nd part
                {
                    remain_frame[i]--;
                    continue;
                }
                if (remain_frame[i] < 0)                       //Cooldown 1st part
                {
                    if (++remain_frame[i] == 0)                //Blinking
                        remain_frame[i] = getrand % 3 == 0 ? 36 : 30;
                    continue;
                }
                if ((int)(getrand & 0x7F) == 0)                //Not Blinking
                    remain_frame[i] = -5;
            }
            if (raining && (phase = !phase))
                cnt += frameshift(2);
            return cnt;
        }

        private int[] Delay_M = new[] { 407, 412 };
        private int[] Delay_F = new[] { 402, 407 };
        private int fidget() // For more precise timeline
        {
            int delay1 = (int)(getrand % 90); // Random delay
            int delay2 = IsBoy ? Delay_M[getrand & 1] : Delay_F[getrand & 1]; // Decide movement type non-jump/jump
            return delay1 + delay2;
        }

        public int frameshift(int n)
        {
            for (int i = 0; i < n; i++)
                sfmt.Next();
            return n;
        }

        public void CopyTo(ModelStatus st)
        {
            st.remain_frame = (int[])remain_frame.Clone();
            st.phase = phase;
        }

        public ModelStatus Clone()
        {
            ModelStatus st = new ModelStatus(Modelnumber, sfmt);
            CopyTo(st);
            st.raining = raining;
            st.fidget_cd = fidget_cd;
            st.IsBoy = IsBoy;
            return st;
        }
    }
}